/*******************************************************************************
*    Wecker 1.0.                                                               *
*    Copyright (C) 2010-averyfarwaydate Luca Bertoncello                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    Circuit by Luca Bertoncello and Tim Rettmann, Layout by Tim Rettmann      *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "inc/wecker.h"
#include "inc/tools.h"
#include "inc/dogm.h"
#include "inc/dcf77.h"
#include "inc/time.h"
#include "inc/alarm.h"
#include "resources/dict.h"

/**
* So viele IRQs werden jede Sekunde ausgelöst.
* Für optimale Genauigkeit muss
* IRQS_PER_SECOND ein Teiler von F_CPU sein
* und IRQS_PER_SECOND ein Vielfaches von 100.
* Ausserdem muss gelten F_CPU / IRQS_PER_SECOND <= 65536
*/
#define IRQS_PER_SECOND   1000   /* 1 ms */

/**
* Anzahl IRQs pro 10 Millisekunden
*/
#define IRQS_PER_10MS     (IRQS_PER_SECOND / 100)

/**
* Gültigkeitsprüfung.
* Bei ungeeigneten Werten gibt es einen Compilerfehler
*/
#if (F_CPU/IRQS_PER_SECOND > 65536) || (IRQS_PER_10MS < 1) || (IRQS_PER_10MS > 255)
#   error Diese Werte für F_CPU und IRQS_PER_SECOND
#   error sind ausserhalb des gueltigen Bereichs!
#endif

/**
* Compiler-Warnung falls die Genauigkeit nicht optimal ist.
* Wenn das nervt für deine Werte, einfach löschen :-)
*/
#if (F_CPU % IRQS_PER_SECOND != 0) || (IRQS_PER_SECOND % 100 != 0)
#   warning Das Programm arbeitet nicht mit optimaler Genauigkeit.
#endif

#if !defined (TCNT1H)
#error Dieser Controller hat keinen 16-Bit Timer1!
#endif // TCNT1H

/**
* Die Interrupt Service Routine (ISR).
* Die Variabel "uptime" wird erhöht
*/
SIGNAL (TIM1_COMPA_vect)
{
  uint8_t max = 60;

  uptime++;

  if((uptime % 1000) == 0)
  {
    setToDisplay();
    currentTime.seconds++;
    if(currentTime.minutes == 59 && currentTime.leapSecond == 1) // Schaltsekunde
      max++;
    if(currentTime.seconds == max)
    {
      currentTime.seconds = 0;
      currentTime.minutes++;
      if(currentTime.minutes == 60)
      {
        currentTime.leapSecond = 0;   // Schaltsekunde zurücksetzen
        currentTime.minutes = 0;
        if(currentTime.dstChange == 1)
        {
          if(currentTime.dst == SUMMERTIME)
          {
            currentTime.dst = WINTERTIME;
            currentTime.hours--;
          }
          else
          {
            currentTime.dst = SUMMERTIME;
            currentTime.hours++;
          }
        }
        currentTime.hours++;
        if(currentTime.hours == 24)
        {
          currentTime.hours = 0;
          changeDay(TYPE_CURRENTTIME, DIRECTION_UP);
        }
      }
    }
  }
}

/**
* Initialisiert den Timer
*/
void dcf77InitTimer()
{
  uptime = 0L;
  toDisplay = TRUE;

  // Timer1: keine PWM
  TCCR1A = 0;

  // Timer1 ist Zähler: Clear Timer on Compare Match (CTC, Mode #4)
  // Timer1 läuft mit vollem MCU-Takt: Prescale = 1
#if defined (CTC1) && !defined (WGM12)
  TCCR1B = (1 << CTC1)  | (1 << CS10);
#elif !defined (CTC1) && defined (WGM12)
  TCCR1B = (1 << WGM12) | (1 << CS10);
#else
#error Keine Ahnung, wie Timer1 für diesen AVR zu initialisieren ist!
#endif

  // OutputCompare für gewünschte Timer1 Frequenz
  // TCNT1 zählt immer 0...OCR1A, 0...OCR1A, ...
  // Beim überlauf OCR1A -> OCR1A+1 wird TCNT1=0 gesetzt und im nächsten
  // MCU-Takt eine IRQ erzeugt.
  OCR1A = (unsigned short) ((unsigned long) F_CPU / IRQS_PER_SECOND-1);

  // OutputCompareA-Interrupt für Timer1 aktivieren
#if defined (TIMSK1)
  TIMSK1 |= (1 << OCIE1A);
#elif defined (TIMSK)
  TIMSK  |= (1 << OCIE1A);
#else
#error Keine Ahnung, wie IRQs für diesen AVR zu initialisieren sind!
#endif
}

/**
 * Initialisiert das DCF77-Modul
 */
void dcf77Init()
{
  DCF77_UB_DDR |= _BV(DCF77_UB);
  DCF77_SIGNAL_DDR &= ~_BV(DCF77_SIGNAL);
  currentTime.seconds = currentTime.minutes = currentTime.hours = \
    currentTime.weekDay = currentTime.dst = currentTime.leapSecond = currentTime.dstChange = 0;
  currentTime.month = currentTime.day = 1;
  currentTime.year = 10;
  lastSyncTime.seconds = lastSyncTime.minutes = lastSyncTime.hours = \
    lastSyncTime.weekDay = lastSyncTime.dst = lastSyncTime.leapSecond = lastSyncTime.dstChange = 0;
  lastSyncTime.month = lastSyncTime.day = 1;
  lastSyncTime.year = 10;
  doSync = FALSE;
  setTimezone(DEFAULTTIMEZONE);
  hasToSync = displayHoliday = FALSE;
  hourToSync = 1;   // Bis zum ersten erfolgreichen Sync, ein Versuch jede Stunde
}

/**
 * Holt die aktuelle Zeit vom DCF77
 *
 * @param bool first              TRUE wenn der Wecker gerade eingeschaltet worden ist
 * @return bool                   TRUE wenn die Zeit synchron ist
 */
bool dcf77GetTime(bool first)
{
  uint8_t bit, l, parity;
  volatile unsigned long start, len, wait, startSync;
  char d = '*';
  uint8_t canStart = 0, nSec = 0, stop;
  uint8_t data[60];
  struct dcf77Time newTime;

  doSync = TRUE;
  newTime.seconds = newTime.minutes = newTime.hours = \
    newTime.year = newTime.month = newTime.day = newTime.weekDay = \
    newTime.dst = newTime.leapSecond = newTime.dstChange = 0;
  dcf77TurnOn();
  bit = 255;
  start = stop = 0;
  wait = 0;
  valid = FALSE;
  startSync = getUptime();
  while((stop == 0) && (getUptime() < (startSync + MAXSYNCTIME)))
  {
    displayCurrentTime();
    if(bit_is_set(DCF77_SIGNAL_PIN, DCF77_SIGNAL))
    {
      if(start == 0)
      {
        lcdGotoXY(15, 0);
        lcdPutc(d);
        start = getUptime();
      }
      if(wait != 0)
      {
        if((getUptime() - wait) > 1500)
        {
          canStart = 1;
        }
        wait = 0;
      }
    }
    else
    {
      if(wait == 0)
        wait = getUptime();
      if(start != 0)
      {
        len = getUptime() - start;
        start = 0;
        lcdGotoXY(15, 0);
        lcdPutc(' ');
        if(len < 140)
          bit = 0;
        else
          bit = 1;
        if(canStart == 2)
        { // Gut, wir empfangen die Daten...
          nSec++;
          lcdGotoXY(0, 2);
          lcdPuts(RECEIVINGDATA);
          if(nSec >= 1 && nSec <= 14)
          { // (Wetterdaten von Meteotime verschlüsselt)
#ifdef DEBUG
            d = '1';
#endif
          }
          if(nSec == 15)
          { // 0=Normalantenne, 1=Reserveantenne
#ifdef DEBUG
            d = '2';
#endif
          }
          if(nSec == 16)
          { // Umstellung zwischen Sommer- bzw. Winterzeit
            if(bit == 1)
              newTime.dstChange = 1;
            else
              newTime.dstChange = 0;
#ifdef DEBUG
            d = '3';
#endif
          }
          if(nSec == 17)
          {
            if(bit == 1)
              newTime.dst = SUMMERTIME;
#ifdef DEBUG
            d = '4';
#endif
          }
          if(nSec == 18)
          {
            if(bit == 1)
              newTime.dst = WINTERTIME;
#ifdef DEBUG
            d = '5';
#endif
          }
          if(nSec == 19)
          { // 1=Am Ende dieser Stunde wird eine Schaltsekunde eingefügt
            if(bit == 1)
              newTime.leapSecond = 1;
            else
              newTime.leapSecond = 0;
#ifdef DEBUG
            d = '6';
#endif
          }
          if(nSec == 20)
          { // Dieses Bit sollte immer 1 sein. Wenn nicht, dann ist einen Empfangsfehler
            if(bit == 0)
            {
              canStart = 0; // nächster Versuch...
              lcdGotoXY(0, 2);
              lcdPuts(FAILEDTOSYNC);
              d = '*';
            }
#ifdef DEBUG
            d = '7';
#endif
          }
          if(canStart != 0 && nSec >= 21 && nSec <= 27)
          { // Minuten
            data[nSec - 21] = bit;
#ifdef DEBUG
            d = '8';
#endif
          }
          if(canStart != 0 && nSec == 28)
          { // Minutenparität
            for(parity = l = 0; l < 7; l++)
              if(data[l] == 1)
                parity++;
            if((parity % 2) == bit)
            {
              newTime.minutes = data[0] * 1 + \
                                data[1] * 2 + \
                                data[2] * 4 + \
                                data[3] * 8 + \
                                data[4] * 10 + \
                                data[5] * 20 + \
                                data[6] * 40;
            }
#ifdef DEBUG
            d = '9';
#endif
          }
          if(canStart != 0 && nSec >= 29 && nSec <= 34)
          { // Stunden
            data[nSec - 29] = bit;
#ifdef DEBUG
            d = 'A';
#endif
          }
          if(canStart != 0 && nSec == 35)
          { // Stundenparität
            for(parity = l = 0; l < 6; l++)
              if(data[l] == 1)
                parity++;
            if((parity % 2) == bit)
            {
              newTime.hours = data[0] * 1 + \
                              data[1] * 2 + \
                              data[2] * 4 + \
                              data[3] * 8 + \
                              data[4] * 10 + \
                              data[5] * 20;
            }
#ifdef DEBUG
            d = 'B';
#endif
          }
          if(canStart != 0 && nSec >= 36 && nSec <= 41)
          { // Tag
            data[nSec - 36] = bit;
#ifdef DEBUG
            d = 'C';
#endif
          }
          if(canStart != 0 && nSec >= 42 && nSec <= 44)
          { // Wochentag
            data[nSec - 36] = bit;
#ifdef DEBUG
            d = 'D';
#endif
          }
          if(canStart != 0 && nSec >= 45 && nSec <= 49)
          { // Monat
            data[nSec - 36] = bit;
#ifdef DEBUG
            d = 'E';
#endif
          }
          if(canStart != 0 && nSec >= 50 && nSec <= 57)
          { // Jahr
            data[nSec - 36] = bit;
#ifdef DEBUG
            d = 'F';
#endif
          }
          if(canStart != 0 && nSec == 58)
          { // Parität
#ifdef DEBUG
            d = '!';
#endif
            for(parity = l = 0; l < 22; l++)
              if(data[l] == 1)
                parity++;
            if((parity % 2) == bit)
            {
              newTime.day = data[0] * 1 + \
                            data[1] * 2 + \
                            data[2] * 4 + \
                            data[3] * 8 + \
                            data[4] * 10 + \
                            data[5] * 20;
              newTime.weekDay = data[6] * 1 + \
                                data[7] * 2 + \
                                data[8] * 4;
              newTime.month = data[9] * 1 + \
                              data[10] * 2 + \
                              data[11] * 4 + \
                              data[12] * 8 + \
                              data[13] * 10;
              newTime.year = data[14] * 1 + \
                             data[15] * 2 + \
                             data[16] * 4 + \
                             data[17] * 8 + \
                             data[18] * 10 + \
                             data[19] * 20 + \
                             data[20] * 40 + \
                             data[21] * 80;
              canStart = 0;
              if(areDataValid(newTime, first) == TRUE)
              {
                valid = TRUE;
                lcdGotoXY(0, 2);
                lcdPuts(RECEIVEDDATA);
              }
              else
              {
                lcdGotoXY(0, 2);
                lcdPuts(FAILEDTOSYNC);
              }
            }
          }
        }
        if(canStart == 1)
        {
          if(bit == 0)
          {
            if(valid == TRUE)
            {
// Der Zähler der Millisekunden wird auf 0 gesetzt. Dann sind die Sekunden auch mit dem DCF77 synchron.
              uptime = 0L;
              currentTime.minutes = newTime.minutes;
              currentTime.hours = newTime.hours;
              currentTime.year = newTime.year;
              currentTime.month = newTime.month;
              currentTime.day = newTime.day;
              if(newTime.weekDay == 7)
                currentTime.weekDay = 0;
              else
                currentTime.weekDay = newTime.weekDay;
// Zeit nach der Zeitzone anpassen
              currentTime.hours += timezone - DEFAULTTIMEZONE;
              if(currentTime.hours > 23)
              {
                currentTime.hours -= 24;
                changeDay(TYPE_CURRENTTIME, DIRECTION_UP);
              }
              if(currentTime.hours < 0)
              {
                currentTime.hours += 24;
                changeDay(TYPE_CURRENTTIME, DIRECTION_DOWN);
              }
              lastSyncTime.seconds = currentTime.seconds = 0;
              lastSyncTime.minutes = currentTime.minutes;
              lastSyncTime.hours = currentTime.hours;
              lastSyncTime.year = currentTime.year;
              lastSyncTime.month = currentTime.month;
              lastSyncTime.day = currentTime.day;
              lastSyncTime.weekDay = currentTime.weekDay;
              lastSyncTime.dst = currentTime.dst = newTime.dst;
              lastSyncTime.leapSecond = currentTime.leapSecond = newTime.leapSecond;
              lastSyncTime.dstChange = currentTime.dstChange = newTime.dstChange;
              stop = 1;
              lcdGotoXY(0, 2);
              lcdPuts(CLEARLINE);
              lcdGotoXY(15, 0);
              lcdPutc(' ');
              doneSync();
              hourToSync = 4;   // Ab jetzt wird die Zeit alle 4 Stunden synchronisiert
            }
            else
            {
              canStart = 2;
              nSec = 0;
              bzero(data, 60);
            }
          }
          else
            canStart = 0; // Ungültig! Bit bei Sekunde "0" ist IMMER 0
        }
        if(canStart == 0 && valid == FALSE)
        {
          lcdGotoXY(0, 2);
          lcdPuts(WAITFORSTART);
          d = '*';
        }
      }
    }
  }
  dcf77TurnOff();
  doSync = FALSE;
  if(valid == FALSE)
  {
    displayError(FAILEDTOSYNC);
    doneSync();
  }
  return valid;
}

/**
 * Prüft die empfangenen Daten und gibt zurück, ob sie gültig sind
 * (0 <= Minuten <= 59, 0 <= Stunde <= 23, usw.)
 *
 * @param struct dcf77Time time   die empfangene Zeit
 * @param bool first              TRUE wenn der Wecker gerade eingeschaltet worden ist
 * @return bool                   TRUE wenn die Daten gültig sind
 */
bool areDataValid(struct dcf77Time time, bool first)
{
  uint8_t days[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

  if(time.minutes > 59)
    return FALSE;
  if(time.hours > 23)
    return FALSE;
  if(time.month == 0 || time.month > 12)
    return FALSE;
  if(((time.year % 4) == 0 && (time.year % 100) != 0) || ((time.year % 400) == 0))
    days[1] = 29;
  if(time.day == 0 || time.day > days[time.month - 1])
    return FALSE;
  if(time.year < 10) // Projekt erst 2010 angefangen. Wenn Jahr kleiner ist, ist ein Fehler!
    return FALSE;

  if(first) // wenn der Wecker gerade eingeschaltet worden ist, jedes Datum wird akzeptiert
    return TRUE;

// Das empfangene Datum MUSS gleich das aktuelle.
// Ansonsten ist es ein Empfangsfehler
// Hintergrund: oft passiert, daß ein falsches Datum empfangen wird.
// Das Prüfbit bei dem Datum ist oft nicht reichlich, es kann aber nicht sein,
// daß die interne Zeit so unterschiedlich von der empfangenen ist, denn die
// interne Uhr hat ein Fehler von weniger als eine Sekunde pro Tag.
  if((time.weekDay != currentTime.weekDay) &&
     !(time.weekDay == 7 && currentTime.weekDay == 0)) // Sonntag ist 7, aber intern wird als 0 gespeichert!!!
    return FALSE;
  if(time.day != currentTime.day ||
     time.month != currentTime.month ||
     time.year != currentTime.year)
    return FALSE;

// Manchmal passiert, dass eine falsche Zeit empfangen wird.
// Die Stunde darf maximal eine Stunden hin- und her wandern (wegen Sommer-/Winterzeit). Mehr nicht
  if(abs(time.hours - currentTime.hours) > 1)
    return FALSE;

  return TRUE;
}

/**
 * Wechselt zum vorigen, bzw. nächsten Tag.
 * Diese Funtion wird jeden Tag aufgerufen, wenn die Zeit zwischen 23:59 und 00:00
 * wechselt, sowie wenn, wegen Zeitzone, der Tag anderes als von DCF77 ist
 *
 * @param uint8_t type            TYPE_CURRENTTIME wenn die aktuelle Zeit geändert werden soll
 *                                TYPE_ALARMTIME wenn die Weckzeit geändert werden soll
 *                                TYPE_CHANGEALARMTIME wenn die vorläufige Weckzeit geändert werden soll
 * @param uint8_t direction       DIRECTION_UP wenn zum nächsten Tag gewechselt werden soll,
 *                                DIRECTION_DOWN wenn zum vorigen Tag gewechselt werden soll
 *                                ACHTUNG! DIRECTION_DOWN arbeitet nur mit TYPE_CURRENTTIME und TYPE_CHANGEALARMTIME (kein Sinn die Weckzeit dem vorigen Tag zu setzen)
 */
void changeDay(uint8_t type, uint8_t direction)
{
  uint8_t days[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

// currentTime und alarmTime sind immer in demselben Jahr.
// Ausnahmefall ist zwischen Silvester und Neujahr, aber hier wird nur berechnet, ob
// das Jahr ein Schaltjahr ist. Also, nicht wichtig für die Alarme
  if(((currentTime.year % 4) == 0 && (currentTime.year % 100) != 0) || ((currentTime.year % 400) == 0))
    days[1] = 29;
  if(direction == DIRECTION_UP)
  { // wechselt zum nächsten Tag
    if(type == TYPE_CURRENTTIME)
    {
      currentTime.day++;
      currentTime.weekDay++;
      if(currentTime.weekDay == 7)
        currentTime.weekDay = 0;
      if(currentTime.day > days[currentTime.month - 1])
      {
        currentTime.day = 1;
        currentTime.month++;
        if(currentTime.month > 12)
        {
          currentTime.month = 1;
          currentTime.year++;
        }
      }
    }
    else if(type == TYPE_CHANGEALARMTIME)
    {
      changeAlarmTime.day++;
      changeAlarmTime.weekDay++;
      if(changeAlarmTime.weekDay == 7)
        changeAlarmTime.weekDay = 0;
      if(changeAlarmTime.day > days[changeAlarmTime.month - 1])
      {
        changeAlarmTime.day = 1;
        changeAlarmTime.month++;
        if(changeAlarmTime.month > 12)
        {
          changeAlarmTime.month = 1;
          changeAlarmTime.year++;
        }
      }
    }
    else
    {
      alarmTime.day++;
      alarmTime.weekDay++;
      if(alarmTime.weekDay == 7)
        alarmTime.weekDay = 0;
      if(alarmTime.day > days[alarmTime.month - 1])
      {
        alarmTime.day = 1;
        alarmTime.month++;
        if(alarmTime.month > 12)
        {
          alarmTime.month = 1;
          alarmTime.year++;
        }
      }
    }
  }
  if(direction == DIRECTION_DOWN)
  { // wechselt zum vorigen Tag
    if(type == TYPE_CURRENTTIME)
    {
      currentTime.day--;
      if(currentTime.weekDay == 0)
        currentTime.weekDay = 6;
      else
        currentTime.weekDay--;
      if(currentTime.day == 0)
      {
        currentTime.month--;
        if(currentTime.month == 0)
        {
          currentTime.month = 12;
          currentTime.year--;
        }
        currentTime.day = days[currentTime.month - 1];
      }
    }
    else if(type == TYPE_CHANGEALARMTIME)
    {
      changeAlarmTime.day--;
      if(changeAlarmTime.weekDay == 0)
        changeAlarmTime.weekDay = 6;
      else
        changeAlarmTime.weekDay--;
      if(changeAlarmTime.day == 0)
      {
        changeAlarmTime.month--;
        if(changeAlarmTime.month == 0)
        {
          changeAlarmTime.month = 12;
          changeAlarmTime.year--;
        }
        changeAlarmTime.day = days[changeAlarmTime.month - 1];
      }
    }
  }
}
