/*******************************************************************************
*    Wecker 1.0.                                                               *
*    Copyright (C) 2010-averyfarwaydate Luca Bertoncello                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    Circuit by Luca Bertoncello and Tim Rettmann, Layout by Tim Rettmann      *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#ifndef DCF77_H_
#define DCF77_H_

#include <avr/io.h>

#define SUMMERTIME            0
#define WINTERTIME            1
struct dcf77Time
{
  uint8_t seconds;
  uint8_t minutes;
  int8_t hours;       // Die Stunde kann auch negativ sein, nach der Korrektur zum Zeitzone und bevor die Berüchsichtigung des Tages
  uint8_t year;       // Jahr fängt ab 2000 an. Also, hier wird nur der Anzahl der Jahr - 2000 gespeichert
  uint8_t month;
  uint8_t day;
  uint8_t weekDay;
  uint8_t dst;        // 0 = Sommerzeit, 1 = Winterzeit
  uint8_t leapSecond; // Schaltsekunde
  uint8_t dstChange;  // 1 = am Ende der Stunde wird die Zeit gewechselt
};

#ifdef SOURCE_dcf77
/**
* Millisekunden vom Einschalten/Reset
*/
volatile unsigned long uptime, displayErrorTime;
struct dcf77Time currentTime, lastSyncTime;
bool doSync, valid, toDisplay, hasToSync, displayHoliday;
int8_t timezone;
uint8_t hourToSync;
#else
extern volatile unsigned long uptime, displayErrorTime;
extern struct dcf77Time currentTime, lastSyncTime;
extern bool doSync, valid, toDisplay, hasToSync, displayHoliday;
extern int8_t timezone;
extern uint8_t hourToSync;
#endif

#define DCF77_SIGNAL_PORT     PORTB
#define DCF77_SIGNAL_DDR      DDRB
#define DCF77_SIGNAL_PIN      PINB
#define DCF77_SIGNAL          1

#define DCF77_UB_PORT         PORTB
#define DCF77_UB_DDR          DDRB
#define DCF77_UB_PIN          PINB
#define DCF77_UB              0

#define DIRECTION_UP          0  // Wechselt zum nächsten Tag
#define DIRECTION_DOWN        1  // Wechselt zum vorigen Tag
#define TYPE_CURRENTTIME      0  // Es wird die aktuelle Zeit verwendet
#define TYPE_ALARMTIME        1  // Es wird die Weckzeit verwendet
#define TYPE_CHANGEALARMTIME  2  // Es wird die vorläufige Weckzeit verwendet

/**
 * Gibt der Signal, daß die Zeit auf dem Display zu aktualisieren ist
 */
#define setToDisplay()              toDisplay = TRUE;

/**
 * Zeit wurde aktualisiert
 */
#define clearToDisplay()            toDisplay = FALSE;

/**
 * Gibt zurück, ob die Zeit auf dem Display zu aktualisieren ist
 */
#define hasToDisplay()              toDisplay == TRUE

/**
 * Startet die Synchronisierung
 */
#define startSync()                 hasToSync = TRUE;

/**
 * Signalisiert, daß die Synchronisierung abgeschloßen ist
 */
#define doneSync()                  hasToSync = FALSE;

/**
 * Gibt zurück, ob die Synchronisierung der Zeit gestartet werden soll
 */
#define hasToStartSync()            (hasToSync == TRUE)

/**
 * Setzt die Zeitzone
 *
 * @param int8_t __timezone       Die Zeitzone
 */
#define setTimezone(__timezone)     timezone = __timezone;

/**
 * Initialisiert das DCF77-Modul
 */
void dcf77Init(void);

/**
 * Holt die aktuelle Zeit vom DCF77
 *
 * @param bool first              TRUE wenn der Wecker gerade eingeschaltet worden ist
 * @return bool                   TRUE wenn die Zeit synchron ist
 */
bool dcf77GetTime(bool first);

/**
 * Schaltet das DCF77-Modul ein
 */
#define dcf77TurnOn()               DCF77_UB_PORT |= _BV(DCF77_UB);

/**
 * Schaltet das DCF77-Modul aus
 */
#define dcf77TurnOff()              DCF77_UB_PORT &= ~_BV(DCF77_UB);

/**
* Initialisiert den Timer
*/
void dcf77InitTimer(void);

/**
* Gibt den Uptime (Millisekunden seit dem letzten Einschalten/Reset)
* zurück
*
* @return long                    Die Millisekunden seit dem letzten Einschalten/Reset
*/
#define getUptime()                 uptime

/**
 * Gibt zurück, ob gerade einen DCF-Synchronisierung läuft
 */
#define isSynchonizing()            doSync

/**
 * Prüft die empfangenen Daten und gibt zurück, ob sie gültig sind
 * (0 <= Minuten <= 59, 0 <= Stunde <= 23, usw.)
 *
 * @param struct dcf77Time time   die empfangene Zeit
 * @param bool first              TRUE wenn der Wecker gerade eingeschaltet worden ist
 * @return bool                   TRUE wenn die Daten gültig sind
 */
bool areDataValid(struct dcf77Time time, bool first);

/**
 * Wechselt zum vorigen, bzw. nächsten Tag.
 * Diese Funtion wird jeden Tag aufgerufen, wenn die Zeit zwischen 23:59 und 00:00
 * wechselt, sowie wenn, wegen Zeitzone, der Tag anderes als von DCF77 ist
 *
 * @param uint8_t type            TYPE_CURRENTTIME wenn die aktuelle Zeit geändert werden soll
 *                                TYPE_ALARMTIME wenn die Weckzeit geändert werden soll
 *                                TYPE_CHANGEALARMTIME wenn die vorläufige Weckzeit geändert werden soll
 * @param uint8_t direction       DIRECTION_UP wenn zum nächsten Tag gewechselt werden soll,
 *                                DIRECTION_DOWN wenn zum vorigen Tag gewechselt werden soll
 *                                ACHTUNG! DIRECTION_DOWN arbeitet nur mit TYPE_CURRENTTIME und TYPE_CHANGEALARMTIME (kein Sinn die Weckzeit dem vorigen Tag zu setzen)
 */
void changeDay(uint8_t type, uint8_t direction);

#endif /* DCF77_H_ */
